/*--------------------------------------------------------*/
/* DEFINE										*/
/*--------------------------------------------------------*/
#define		DF_PORT_DATA		0x378			/* parallel port data register address */
#define		DF_PORT_STATUS   0x379			/* parallel port status register address */
#define		DF_PORT_CONTROL  0x37A			/* parallel port control register address */
#define		DF_IRQ				7				/* IRQ number for parallel port */
#define		DF_MAJOR			254				/* major number for pio */
#define		DF_MINOR			0				/* minor number for pio */

//#define		DF_DEBUG_PARREAD_INIT			/* debug for parread_init() */
//#define		DF_DEBUG_PARREAD_CLEANUP		/* debug for parread_cleanup() */
//#define		DF_DEBUG_PARREAD_OPEN			/* debug for parread_open() */
//#define		DF_DEBUG_PARREAD_RELEASE		/* debug for parread_release() */
//#define		DF_DEBUG_PARREAD_INTERRUPT		/* debug for parread_interrupt() */
//#define 		DF_DEBUG_PARREAD_IN_SPACE		/* debug for parread_in_space() */
//#define		DF_DEBUG_PARREAD_INCR_BP		/* debug for parread_incr_bp() */
//#define		DF_DEBUG_PARREAD_READ			/* debug for parread_read() */
/*--------------------------------------------------------*/
/* INCLUDE									   */
/*--------------------------------------------------------*/
#include	<linux/module.h>			/* THIS_MODULE */
#include	<linux/ioport.h>			/* request_region() */
#include	<linux/poll.h>			/* */
#include	<linux/cdev.h>			/* cdev_alloc(), cdev_add(), cdev_del() */
#include	<asm/io.h>					/* inb_p(), outb() */
#include	<rtai.h>					/* rtai */
/*--------------------------------------------------------*/
/* PROTOTYPE											  */
/*--------------------------------------------------------*/
static int __init parread_init(void);												/* Initialize module */
static void parread_cleanup(void);													/* Clean up module */
static int parread_open(struct inode *,struct file *);							/* Open the device */
static int parread_release(struct inode *,struct file *);						/* Close the device */
static void parread_interrupt(void);												/* IRQ Handler */
static unsigned long parread_in_space(unsigned long,unsigned long);  		/* Check the ring buffer space */
static void parread_incr_bp(volatile unsigned long *,unsigned long); 		/* Increase the ring buffer pointer */
static ssize_t parread_read(struct file *,char __user *,size_t,loff_t *);	/* Read the ring buffer */
module_init(parread_init);															/* Entry the insmod module */
module_exit(parread_cleanup);														/* Entry the rmmod module */
/*--------------------------------------------------------*/
/* GLOVALS												*/
/*--------------------------------------------------------*/
static dev_t						G_dev_t;					/* Device Number */
static struct cdev *				G_cdev;					/* cdev for file operations */
static unsigned char				G_ucStatus;				/* Init status */
volatile static int				G_siActive;				/* pio active flag */
static unsigned long				G_ulRB_Base;				/* Ring buffer base address */
static volatile unsigned long	G_ulRB_Head;				/* Ring buffer space header */
static volatile unsigned long	G_ulRB_Tail;				/* Ring buffer data header */
static struct file_operations	G_fops =
{
	.read =	   parread_read,
	.open =	   parread_open,
	.release = parread_release,
};
MODULE_LICENSE("GPL");
/*--------------------------------------------------------*/
/*--------------------------------------------------------*/
/*--------------------------------------------------------*/
/*--------------------------------------------------------*/
/* Initialize module									  */
/*--------------------------------------------------------*/
static int __init parread_init(void)
{
	int wk_siRet = 0;																			/* returning variable */

	do
	{
		G_ucStatus = 0x00;																	/* Initialize the status */
		G_siActive = 0;																		/* IRQ is ignore */

		if(request_region(DF_PORT_DATA,3,"parread") != 0)								/* Get the I/O port[0x378-0x37A] */
		{
			G_ucStatus |= 0x01;																/* Update status */
		}
		else
		{
			printk(KERN_INFO "parread: can't get I/O port address 0x378-0x37A\n");	/* Output error message */
			wk_siRet = -ENODEV;																/* Set NG */
			break;																				/* cancel the init */
		}

		G_dev_t = MKDEV(DF_MAJOR,DF_MINOR);												/* Get the device number from major number and minor number */
		if(register_chrdev_region(G_dev_t,1,"parread") >= 0)							/* Entry the device number*/
		{
			G_ucStatus |= 0x02;																/* Update status */
		}
		else
		{
			printk(KERN_INFO "parread: can't entry the device number\n");				/* Output error message */
			wk_siRet = -ENODEV;																/* Set NG */
			break;																				/* cancel the init */
		}

		G_cdev = cdev_alloc();																/* Get the cdev area */
		G_cdev-> owner = THIS_MODULE;														/* Set the owner */
		G_cdev-> ops = &G_fops;																/* entry the fops */
		if(cdev_add(G_cdev,G_dev_t,1) >= 0)												/* add the cdev */
		{
			G_ucStatus |= 0x04;																/* Update status */
		}
		else
		{
			printk(KERN_INFO "parread: can't add the device\n");						/* Output error message */
			wk_siRet = -ENODEV;																/* Set NG */
			break;																				/* cancel the init */
		}

		if(rt_request_global_irq(DF_IRQ,parread_interrupt) == 0)						/* entry the irq handler */
		{
			rt_enable_irq(DF_IRQ);															/* enable irq7 */
			G_ucStatus |= 0x08;																/* Update status */
		}
		else
		{
			printk(KERN_INFO "parread: can't entry the irq.\n");						/* output error message */
			wk_siRet = -ENODEV;																/* Set NG */
			break;																				/* cancel the init */
		}

		G_ulRB_Base = __get_free_pages(GFP_KERNEL, 0);									/* Get the memory for the ring buffer */
		G_ulRB_Head = G_ulRB_Tail = G_ulRB_Base;											/* Init pointers */
		G_ucStatus |= 0x10;																	/* Update status */

		/* MODE = Hi(Input), IRQ = Hi(Enable), SELECT = Lo(nc), INIT = Lo(nc) AUTOLF = Lo(nc) STROBE = Lo(nc) */
		outb(0x30,DF_PORT_CONTROL);
	}while(0);

	if(wk_siRet != 0)																			/* error? */
	{
		parread_cleanup();																	/* clean up */
	}

#ifdef DF_DEBUG_PARREAD_INIT
	printk(KERN_INFO "parread: parread_init: G_ulRB_Base = %lX\n",G_ulRB_Base);
	printk(KERN_INFO "parread: parread_init: G_ulRB_Head = %lX\n",G_ulRB_Head);
	printk(KERN_INFO "parread: parread_init: G_ulRB_Tail = %lX\n",G_ulRB_Tail);
	printk(KERN_INFO "parread: parread_init: G_ucStatus = %2X\n",G_ucStatus);
	printk(KERN_INFO "parread: parread_init: G_siActive = %d\n",G_siActive);
#endif
	return wk_siRet;
}
/*--------------------------------------------------------*/
/* Clean up module							        */
/*--------------------------------------------------------*/
static void __exit parread_cleanup(void)
{
	/* free the ring buffer memory */
	if((G_ucStatus & 0x10) != 0)
	{
		G_ucStatus &= 0xEF;
		free_page(G_ulRB_Base);
	}

	/* delete irq */
	if((G_ucStatus & 0x08) != 0)
	{
		G_ucStatus &= 0xF7;
		outb(0x0,DF_PORT_CONTROL);			/* disable the interrupt */
		rt_disable_irq(DF_IRQ);				/* disable irq7 */
		rt_free_global_irq(DF_IRQ);			/* delete irq7 */
	}

	/* delete cdev */
	if((G_ucStatus & 0x04) != 0)
	{
		G_ucStatus &= 0xFB;
		cdev_del(G_cdev);
	}

	/* delete device */
	if((G_ucStatus & 0x02) != 0)
	{
		G_ucStatus &= 0xFD;
		unregister_chrdev_region(G_dev_t,1);
	}

	/* delete port */
	if((G_ucStatus & 0x01) != 0)
	{
		G_ucStatus &= 0xFE;
		release_region(DF_PORT_DATA,3);
	}
#ifdef DF_DEBUG_PARREAD_CLEANUP
	printk(KERN_INFO "parread: parread_cleanup: G_ucStatus = %2X\n",G_ucStatus);
#endif
	return;
}
/*--------------------------------------------------------*/
/* Open the device										*/
/*--------------------------------------------------------*/
static int parread_open(struct inode *inode,struct file *filp)
{
	G_ulRB_Head = G_ulRB_Tail = G_ulRB_Base;			/* Initialize the ring buffer */
	G_siActive = 1;										/* interrupt handler enable */

#ifdef DF_DEBUG_PARREAD_OPEN
	printk(KERN_INFO "parread: parread_open: G_ulRB_Base = %lX\n",G_ulRB_Base);
	printk(KERN_INFO "parread: parread_open: G_ulRB_Head = %lX\n",G_ulRB_Head);
	printk(KERN_INFO "parread: parread_open: G_ulRB_Tail = %lX\n",G_ulRB_Tail);
	printk(KERN_INFO "parread: parread_open: G_siActive = %d\n",G_siActive);
#endif
	return 0;
}
/*--------------------------------------------------------*/
/* Close the device									   */
/*--------------------------------------------------------*/
static int parread_release(struct inode *inode, struct file *filp)
{
	G_siActive = 0;

#ifdef DF_DEBUG_PARREAD_RELEASE
	printk(KERN_INFO "parread: parread_release: G_siActive = %d\n",G_siActive);
#endif
	return 0;
}
/*--------------------------------------------------------*/
/* IRQ handler											*/
/*--------------------------------------------------------*/
static void parread_interrupt(void)
{
	unsigned char *	wk_ucDataAttr;							/* ring buffer space address */
	unsigned char		wk_ucData;									/* reception data */

#ifdef DF_DEBUG_PARREAD_INTERRUPT
	printk(KERN_INFO "parread: parread_interrupt: G_ulRB_Head[before] = %lX\n",G_ulRB_Head);
	printk(KERN_INFO "parread: parread_interrupt: G_ulRB_Tail = %lX\n",G_ulRB_Tail);
	printk(KERN_INFO "parread: parread_interrupt: G_ulRB_Base = %lX\n",G_ulRB_Base);
#endif

	rt_ack_irq(DF_IRQ);											/* ack */
	wk_ucData = (char)inb_p(DF_PORT_DATA);					/* Get the reception data */
	if(G_siActive != 0)									 		/* active? */
	{
		if(parread_in_space(G_ulRB_Head,G_ulRB_Tail) > 0)   /* Is there the space in ring buffer? */
		{
			wk_ucDataAttr = (char *)G_ulRB_Head;				/* Get the ring buffer space address */
			(*wk_ucDataAttr) = wk_ucData;						/* write the pio data to the ring buffer */
			parread_incr_bp(&G_ulRB_Head,1);					/* Update the ring buffer space address */
		}
	}
	else
	{
	}

#ifdef DF_DEBUG_PARREAD_INTERRUPT
	printk(KERN_INFO "parread: parread_interrupt: wk_ucData = %2X\n",wk_ucData);
	printk(KERN_INFO "parread: parread_interrupt: G_ulRB_Head[after] = %lX\n",G_ulRB_Head);
#endif
	return;
}
/*--------------------------------------------------------*/
/* Check the ring buffer space				        */
/*--------------------------------------------------------*/
static unsigned long parread_in_space(unsigned long wk_ulHead,unsigned long wk_ulTail)
{
	unsigned long wk_ulSpace;

#ifdef DF_DEBUG_PARREAD_IN_SPACE
	printk(KERN_INFO "parrread: parread_in_space: wk_ulHead = %lX\n",wk_ulHead);
	printk(KERN_INFO "parrread: prread_in_space: wk_ulTail = %lX\n",wk_ulTail);
#endif

	if(wk_ulHead >= wk_ulTail)				/* The tail pointer(using point) steps over the tie? */
	{
		wk_ulSpace = PAGE_SIZE - (wk_ulHead - wk_ulTail) - 1;
	}
	else
	{
		wk_ulSpace = wk_ulTail - wk_ulHead - 1;
	}

#ifdef DF_DEBUG_PARREAD_IN_SPACE
	printk(KERN_INFO "parread: parread_in_space: wk_ulSpace = %lX\n",wk_ulSpace);
#endif

	return wk_ulSpace;
}
/*--------------------------------------------------------*/
/* Increase the ring buffer pointer			  */
/*--------------------------------------------------------*/
static void parread_incr_bp(volatile unsigned long *wk_ulIndex,unsigned long wk_ulDelta)
{
	unsigned long wk_ulTemp;

#ifdef DF_DEBUG_PARREAD_INCR_BP
	printk(KERN_INFO "parread: parread_incr_bp: *wk_ulIndex[before] = %lX\n",*wk_ulIndex);
	printk(KERN_INFO "parread: parread_incr_bp: wk_ulDelta = %lX\n",wk_ulDelta);
	printk(KERN_INFO "parread: parread_incr_bp: G_ulRB_Base = %lX\n",G_ulRB_Base);
#endif

	/* precondition: wk_ulDelta <= PAGE_SIZE */
	wk_ulTemp = G_ulRB_Base + PAGE_SIZE;						/* not overflow */
	wk_ulTemp = wk_ulTemp -1 - wk_ulDelta;					/* not underflow*/

	if((*wk_ulIndex) > wk_ulTemp)								/* Does it step over the tie? */
	{
		wk_ulTemp = (unsigned long)PAGE_SIZE - wk_ulDelta;	/* not underflow */
		(*wk_ulIndex) = (*wk_ulIndex) - wk_ulTemp;			/* not underflow */
	}
	else
	{
		(*wk_ulIndex) = (*wk_ulIndex) + wk_ulDelta;			/* not overflow */
	}

#ifdef DF_DEBUG_PARREAD_INCR_BP
	printk(KERN_INFO "parread: parread_incr_bp: *wk_ulIndex[after] = %lX\n",*wk_ulIndex);
#endif
	return;
}


/*--------------------------------------------------------*/
/* Read the ring buffer								   */
/*--------------------------------------------------------*/
static ssize_t parread_read(struct file *filp, char __user *wk_siBuf, size_t wk_ulSize, loff_t *f_pos)
{
	unsigned long		wk_ulHead;							/* Ring buffer header */
	unsigned long		wk_ulDataSize;					/* Ring buffer data size */
	unsigned int		wk_uiTie = 0;
	unsigned long		wk_ulTemp;

#ifdef DF_DEBUG_PARREAD_READ
	printk(KERN_INFO "parread: parread_read: G_ulRB_Base = %lX\n",G_ulRB_Base);
	printk(KERN_INFO "parread: parread_read: G_ulRB_Tail[before] = %lX\n",G_ulRB_Tail);
#endif


	wk_ulHead = G_ulRB_Head;								/* G_ulRB_Head is written by the IRQ interrupt */

	if(wk_ulHead != G_ulRB_Tail)
	{
		/* Get the ring buffer data size */
		wk_ulDataSize = PAGE_SIZE -1 - parread_in_space(wk_ulHead,G_ulRB_Tail);

		/* If the request data size is bigger than the ring buffer data size,
		it is changed the ring buffer data size */
		if(wk_ulDataSize < wk_ulSize)
		{
			wk_ulSize = wk_ulDataSize;
		}

		if(wk_ulHead < G_ulRB_Tail)
		{
			wk_ulTemp = G_ulRB_Base + PAGE_SIZE;
			wk_ulTemp = wk_ulTemp - wk_ulSize;
			if(G_ulRB_Tail > wk_ulTemp)
			{
				wk_uiTie = 1;
			}
		}

		if(wk_uiTie == 0)
		{
			if(copy_to_user(wk_siBuf,(char *)G_ulRB_Tail, wk_ulSize))
			{
				return -EFAULT;
			}
		}
		else
		{
			wk_ulTemp = G_ulRB_Base + PAGE_SIZE - G_ulRB_Tail;
			if(copy_to_user(wk_siBuf,(char *)G_ulRB_Tail,wk_ulTemp))
			{
				return -EFAULT;
			}

			if(copy_to_user(wk_siBuf + wk_ulTemp,(char *)G_ulRB_Base,wk_ulSize - wk_ulTemp))
			{
				return -EFAULT;
			}
		}
	}
	else
	{
		wk_ulSize = 0;
	}
	parread_incr_bp(&G_ulRB_Tail,wk_ulSize);

#ifdef DF_DEBUG_PARREAD_READ
	printk(KERN_INFO "parread: parread_read: G_ulRB_Tail[after] = %lX\n",G_ulRB_Tail);
	printk(KERN_INFO "parread: parread_read: wk_uiTie = %d\n",wk_uiTie);
#endif
	return wk_ulSize;
}

